@@ -1,22 +1,23 @@
+# Copyright (c) Jupyter Development Team.
+# Distributed under the terms of the Modified BSD License.
+
 """Base Widget class.  Allows user to create widgets in the back-end that render
 in the IPython notebook front-end.
 """
 
-# Copyright (c) Jupyter Development Team.
-# Distributed under the terms of the Modified BSD License.
-
-
 from contextlib import contextmanager
 import collections
+import sys
 
 from IPython.core.getipython import get_ipython
 from ipykernel.comm import Comm
 from traitlets.config import LoggingConfigurable
-from ipython_genutils.importstring import import_item
-from traitlets import Unicode, Dict, Instance, Bool, List, \
-    CaselessStrEnum, Tuple, CUnicode, Int, Set
-from ipython_genutils.py3compat import string_types
-from .trait_types import Color
+from traitlets.utils.importstring import import_item
+from traitlets import Unicode, Dict, Instance, List, Int, Set, Bytes, observe, default
+from ipython_genutils.py3compat import string_types, PY3
+from IPython.display import display
+
+from .._version import __frontend_version__
 
 
 def _widget_to_json(x, obj):
@@ -44,6 +45,10 @@ widget_serialization = {
     'to_json': _widget_to_json
 }
 
+if PY3:
+    _binary_types = (memoryview, bytes)
+else:
+    _binary_types = (memoryview, buffer)
 
 class CallbackDispatcher(LoggingConfigurable):
     """A structure for registering and running callbacks"""
@@ -97,9 +102,11 @@ def _show_traceback(method):
 
 def register(key=None):
     """Returns a decorator registering a widget class in the widget registry.
-    If no key is provided, the class name is used as a key. A key is
-    provided for each core IPython widget so that the frontend can use
-    this key regardless of the language of the kernel"""
+
+    If no key is provided, the class name is used as a key.
+    A key is provided for each core Jupyter widget so that the frontend can use
+    this key regardless of the language of the kernel.
+    """
     def wrap(widget):
         l = key if key is not None else widget.__module__ + widget.__name__
         Widget.widget_types[l] = widget
@@ -132,29 +139,50 @@ class Widget(LoggingConfigurable):
     @staticmethod
     def handle_comm_opened(comm, msg):
         """Static method, called when a widget is constructed."""
-        widget_class = import_item(str(msg['content']['data']['widget_class']))
+        class_name = str(msg['content']['data']['widget_class'])
+        if class_name in Widget.widget_types:
+            widget_class = Widget.widget_types[class_name]
+        else:
+            widget_class = import_item(class_name)
         widget = widget_class(comm=comm)
 
+    @staticmethod
+    def get_manager_state(drop_defaults=False):
+        return dict(version_major=1, version_minor=0, state={
+            k: {
+                'model_name': Widget.widgets[k]._model_name,
+                'model_module': Widget.widgets[k]._model_module,
+                'model_module_version': Widget.widgets[k]._model_module_version,
+                'state': Widget.widgets[k].get_state(drop_defaults=drop_defaults)
+            } for k in Widget.widgets
+        })
+
+    def get_view_spec(self):
+        return dict(version_major=1, version_minor=0, model_id=self._model_id)
 
     #-------------------------------------------------------------------------
     # Traits
     #-------------------------------------------------------------------------
-    _model_module = Unicode(None, allow_none=True, help="""A requirejs module name
-        in which to find _model_name. If empty, look in the global registry.""", sync=True)
-    _model_name = Unicode('WidgetModel', help="""Name of the backbone model
-        registered in the front-end to create and sync this widget with.""", sync=True)
-    _view_module = Unicode(help="""A requirejs module in which to find _view_name.
-        If empty, look in the global registry.""", sync=True)
-    _view_name = Unicode(None, allow_none=True, help="""Default view registered in the front-end
-        to use to represent the widget.""", sync=True)
+    _model_module = Unicode('jupyter-js-widgets',
+        help="A JavaScript module name in which to find _model_name.").tag(sync=True)
+    _model_name = Unicode('WidgetModel',
+        help="Name of the model object in the front-end.").tag(sync=True)
+    _model_module_version = Unicode('*',
+        help="A semver requirement for the model module version.").tag(sync=True)
+    _view_module = Unicode(None, allow_none=True,
+        help="A JavaScript module in which to find _view_name.").tag(sync=True)
+    _view_name = Unicode(None, allow_none=True,
+        help="Name of the view object.").tag(sync=True)
+    _view_module_version = Unicode('*',
+        help="A semver requirement for the view module version.").tag(sync=True)
     comm = Instance('ipykernel.comm.Comm', allow_none=True)
 
-    msg_throttle = Int(3, sync=True, help="""Maximum number of msgs the
-        front-end can send before receiving an idle msg from the back-end.""")
+    msg_throttle = Int(1, help="""Maximum number of msgs the front-end can send before receiving an idle msg from the back-end.""").tag(sync=True)
 
-    version = Int(0, sync=True, help="""Widget's version""")
     keys = List()
-    def _keys_default(self):
+
+    @default('keys')
+    def _default_keys(self):
         return [name for name in self.traits(sync=True)]
 
     _property_lock = Dict()
@@ -185,15 +213,22 @@ class Widget(LoggingConfigurable):
     def open(self):
         """Open a comm to the frontend if one isn't already open."""
         if self.comm is None:
-            args = dict(target_name='ipython.widget',
-                        data=self.get_state())
+            state, buffer_keys, buffers = self._split_state_buffers(self.get_state())
+
+            args = dict(target_name='jupyter.widget', data=state)
             if self._model_id is not None:
                 args['comm_id'] = self._model_id
-            self.comm = Comm(**args)
 
-    def _comm_changed(self, name, new):
+            self.comm = Comm(**args)
+            if buffers:
+                # FIXME: workaround ipykernel missing binary message support in open-on-init
+                # send state with binary elements as second message
+                self.send_state()
+
+    @observe('comm')
+    def _comm_changed(self, change):
         """Called when the comm is changed."""
-        if new is None:
+        if change['new'] is None:
             return
         self._model_id = self.model_id
 
@@ -221,6 +256,17 @@ class Widget(LoggingConfigurable):
             Widget.widgets.pop(self.model_id, None)
             self.comm.close()
             self.comm = None
+            self._ipython_display_ = None
+
+    def _split_state_buffers(self, state):
+        """Return (state_without_buffers, buffer_keys, buffers) for binary message parts"""
+        buffer_keys, buffers = [], []
+        for k, v in list(state.items()):
+            if isinstance(v, _binary_types):
+                state.pop(k)
+                buffers.append(v)
+                buffer_keys.append(k)
+        return state, buffer_keys, buffers
 
     def send_state(self, key=None):
         """Sends the widget state, or a piece of it, to the front-end.
@@ -231,16 +277,11 @@ class Widget(LoggingConfigurable):
             A single property's name or iterable of property names to sync with the front-end.
         """
         state = self.get_state(key=key)
-        buffer_keys, buffers = [], []
-        for k, v in state.items():
-            if isinstance(v, memoryview):
-                state.pop(k)
-                buffers.append(v)
-                buffer_keys.append(k)
+        state, buffer_keys, buffers = self._split_state_buffers(state)
         msg = {'method': 'update', 'state': state, 'buffers': buffer_keys}
         self._send(msg, buffers=buffers)
 
-    def get_state(self, key=None):
+    def get_state(self, key=None, drop_defaults=False):
         """Gets the widget state, or a piece of it.
 
         Parameters
@@ -263,11 +304,26 @@ class Widget(LoggingConfigurable):
         else:
             raise ValueError("key must be a string, an iterable of keys, or None")
         state = {}
+        traits = self.traits()
         for k in keys:
             to_json = self.trait_metadata(k, 'to_json', self._trait_to_json)
-            state[k] = to_json(getattr(self, k), self)
+            value = to_json(getattr(self, k), self)
+            if not PY3 and isinstance(traits[k], Bytes) and isinstance(value, bytes):
+                value = memoryview(value)
+            if not drop_defaults or not self._compare(value, traits[k].default_value):
+                state[k] = value
         return state
 
+    def _is_numpy(self, x):
+        return x.__class__.__name__ == 'ndarray' and x.__class__.__module__ == 'numpy'
+
+    def _compare(self, a, b):
+        if self._is_numpy(a) or self._is_numpy(b):
+            import numpy as np
+            return np.equal(a, b)
+        else:
+            return a == b
+
     def set_state(self, sync_data):
         """Called when a state is received from the front-end."""
         # The order of these context managers is important. Properties must
@@ -278,11 +334,7 @@ class Widget(LoggingConfigurable):
                 if name in self.keys:
                     from_json = self.trait_metadata(name, 'from_json',
                                                     self._trait_from_json)
-                    # traitlets < 4.1 don't support read-only attributes
-                    if hasattr(self, 'set_trait'):
-                        self.set_trait(name, from_json(sync_data[name], self))
-                    else:
-                        setattr(self, name, from_json(sync_data[name], self))
+                    self.set_trait(name, from_json(sync_data[name], self))
 
     def send(self, content, buffers=None):
         """Sends a custom msg to the widget model in the front-end.
@@ -333,6 +385,18 @@ class Widget(LoggingConfigurable):
                  self.keys.append(name)
                  self.send_state(name)
 
+    def notify_change(self, change):
+        """Called when a property has changed."""
+        # Send the state to the frontend before the user-registered callbacks
+        # are called.
+        name = change['name']
+        if self.comm is not None and self.comm.kernel is not None:
+            # Make sure this isn't information that the front-end just sent us.
+            if name in self.keys and self._should_send_property(name, change['new']):
+                # Send new state to front-end
+                self.send_state(key=name)
+        LoggingConfigurable.notify_change(self, change)
+
     #-------------------------------------------------------------------------
     # Support methods
     #-------------------------------------------------------------------------
@@ -410,21 +474,6 @@ class Widget(LoggingConfigurable):
         """Called when a custom msg is received."""
         self._msg_callbacks(self, content, buffers)
 
-    def _notify_trait(self, name, old_value, new_value):
-        """Called when a property has been changed."""
-        # Trigger default traitlet callback machinery.  This allows any user
-        # registered validation to be processed prior to allowing the widget
-        # machinery to handle the state.
-        LoggingConfigurable._notify_trait(self, name, old_value, new_value)
-
-        # Send the state after the user registered callbacks for trait changes
-        # have all fired (allows for user to validate values).
-        if self.comm is not None and name in self.keys:
-            # Make sure this isn't information that the front-end just sent us.
-            if self._should_send_property(name, new_value):
-                # Send new state to front-end
-                self.send_state(key=name)
-
     def _handle_displayed(self, **kwargs):
         """Called when a view has been displayed for this widget instance"""
         self._display_callbacks(self, **kwargs)
@@ -441,73 +490,59 @@ class Widget(LoggingConfigurable):
 
     def _ipython_display_(self, **kwargs):
         """Called when `IPython.display.display` is called on the widget."""
+        def loud_error(message):
+            self.log.warn(message)
+            sys.stderr.write('%s\n' % message)
+
         # Show view.
         if self._view_name is not None:
+            validated = Widget._version_validated
+
+            # Before the user tries to display a widget, validate that the
+            # widget front-end is what is expected.
+            if validated is None:
+                loud_error('Widget Javascript not detected.  It may not be '
+                           'installed or enabled properly.')
+            elif not validated:
+                msg = ('The installed widget Javascript is the wrong version.'
+                      ' It must satisfy the semver range %s.'%__frontend_version__)
+                if (Widget._version_frontend):
+                    msg += ' The widget Javascript is version %s.'%Widget._version_frontend
+                loud_error(msg)
+
+            # TODO: delete this sending of a comm message when the display statement
+            # below works. Then add a 'text/plain' mimetype to the dictionary below.
             self._send({"method": "display"})
+
+            # The 'application/vnd.jupyter.widget-view+json' mimetype has not been registered yet.
+            # See the registration process and naming convention at
+            # http://tools.ietf.org/html/rfc6838
+            # and the currently registered mimetypes at
+            # http://www.iana.org/assignments/media-types/media-types.xhtml.
+            # We don't have a 'text/plain' entry, so this display message will be
+            # will be invisible in the current notebook.
+            data = {
+                'application/vnd.jupyter.widget-view+json': {
+                    'model_id': self._model_id
+                }
+            }
+            display(data, raw=True)
+
             self._handle_displayed(**kwargs)
 
     def _send(self, msg, buffers=None):
         """Sends a message to the model in the front-end."""
-        self.comm.send(data=msg, buffers=buffers)
+        if self.comm is not None and self.comm.kernel is not None:
+            self.comm.send(data=msg, buffers=buffers)
 
 
-class DOMWidget(Widget):
-    visible = Bool(True, allow_none=True, help="Whether the widget is visible.  False collapses the empty space, while None preserves the empty space.", sync=True)
-    _css = Tuple(sync=True, help="CSS property list: (selector, key, value)")
-    _dom_classes = Tuple(sync=True, help="DOM classes applied to widget.$el.")
-
-    width = CUnicode(sync=True)
-    height = CUnicode(sync=True)
-    # A default padding of 2.5 px makes the widgets look nice when displayed inline.
-    padding = CUnicode(sync=True)
-    margin = CUnicode(sync=True)
-
-    color = Color(None, allow_none=True, sync=True)
-    background_color = Color(None, allow_none=True, sync=True)
-    border_color = Color(None, allow_none=True, sync=True)
-
-    border_width = CUnicode(sync=True)
-    border_radius = CUnicode(sync=True)
-    border_style = CaselessStrEnum(values=[ # http://www.w3schools.com/cssref/pr_border-style.asp
-        'none',
-        'hidden',
-        'dotted',
-        'dashed',
-        'solid',
-        'double',
-        'groove',
-        'ridge',
-        'inset',
-        'outset',
-        'initial',
-        'inherit', ''],
-        default_value='', sync=True)
-
-    font_style = CaselessStrEnum(values=[ # http://www.w3schools.com/cssref/pr_font_font-style.asp
-        'normal',
-        'italic',
-        'oblique',
-        'initial',
-        'inherit', ''],
-        default_value='', sync=True)
-    font_weight = CaselessStrEnum(values=[ # http://www.w3schools.com/cssref/pr_font_weight.asp
-        'normal',
-        'bold',
-        'bolder',
-        'lighter',
-        'initial',
-        'inherit', ''] + list(map(str, range(100,1000,100))),
-        default_value='', sync=True)
-    font_size = CUnicode(sync=True)
-    font_family = Unicode(sync=True)
-
-    def __init__(self, *pargs, **kwargs):
-        super(DOMWidget, self).__init__(*pargs, **kwargs)
-
-        def _validate_border(name, old, new):
-            if new is not None and new != '':
-                if name != 'border_width' and not self.border_width:
-                    self.border_width = 1
-                if name != 'border_style' and self.border_style == '':
-                    self.border_style = 'solid'
-        self.on_trait_change(_validate_border, ['border_width', 'border_style', 'border_color'])
+Widget._version_validated = None
+Widget._version_frontend = None
+def handle_version_comm_opened(comm, msg):
+    """Called when version comm is opened, because the front-end wants to
+    validate the version."""
+    def handle_version_message(msg):
+        Widget._version_validated = msg['content']['data']['validated']
+        Widget._version_frontend = msg['content']['data'].get('frontend_version', '')
+    comm.on_msg(handle_version_message)
+    comm.send({'version': __frontend_version__})
